//------------------------------ $Keywords ----------------------------------
// SelfImage - Windows disk image backup
// SelfImage_MainForm.cpp - Main form
// Copyright 2005, Kurt Fitzner <kfitzner@excelcia.org>
//---------------------------------------------------------------------------
// This file is part of SelfImage.
//
// SelfImage is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// SelfImage is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// VCS: $Version: 0 $ $Revision: 5 $
/*
$History: **** V 0.1 by kfitzner ****
$History: * selfimage_mainform.cpp - 2005-07-30 6:44:58 PM - 2973 Bytes
$History: * selfimage_mainform.dfm - 2005-07-30 5:52:20 PM - 3054 Bytes
$History: * selfimage_mainform.h - 2005-07-30 6:44:58 PM - 1495 Bytes
$History: * selfimage_mainform.ddp - 2005-07-30 6:44:56 PM - 51 Bytes
$History: * Initial check-in
$History: **** V 0.2 by kfitzner ****
$History: * selfimage_mainform.cpp - 2005-08-01 3:11:59 PM - 11269 Bytes
$History: * selfimage_mainform.dfm - 2005-08-01 2:48:30 PM - 6098 Bytes
$History: * selfimage_mainform.h - 2005-08-01 2:38:08 PM - 3569 Bytes
$History: * selfimage_mainform.ddp - 2005-08-01 3:09:38 PM - 51 Bytes
$History: * Initial coding.
$History: **** V 0.3 by kfitzner ****
$History: * selfimage_mainform.cpp - 2005-08-03 4:54:43 AM - 13643 Bytes
$History: * selfimage_mainform.dfm - 2005-08-01 2:48:30 PM - 6098 Bytes
$History: * selfimage_mainform.h - 2005-08-02 9:33:32 PM - 3585 Bytes
$History: * selfimage_mainform.ddp - 2005-08-02 10:21:38 PM - 51 Bytes
$History: * Change writes to use asynchronous I/O
$History: **** V 0.4 by kfitzner ****
$History: * selfimage_mainform.cpp - 2005-08-06 6:29:05 PM - 17290 Bytes
$History: * selfimage_mainform.dfm - 2005-08-06 5:44:58 PM - 6762 Bytes
$History: * selfimage_mainform.h - 2005-08-06 4:39:32 PM - 3586 Bytes
$History: * selfimage_mainform.ddp - 2005-08-06 6:26:46 PM - 51 Bytes
$History: * Add support for unmounted volumes
$History: **** Latest ** V 0.5 by kfitzner ** 2005-08-07 1:43:29 AM ****
$History: * Add version to form caption
*/
//----------------------------- $NoKeywords ---------------------------------

//---------------------------------------------------------------------------
// File Notes:
//---------------------------------------------------------------------------
// 2 Aug 2005 - Kurt Fitzner <kfitzner@excelcia.org>
//
// The main read/write loop of this program uses Windows IO overlapping for
// asynchronous writes.  I have found that this doesn't significantly alter
// the throughput at all, but it does for some reason reduce the CPU
// overhead.
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "SelfImage_MainForm.h"
#include "SelfImage_Exceptions.h"
#include "SelfImage_Utility.h"
#include "TConfiguration.h"
#include "TProgramLog.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma link "JvBevel"
#pragma link "JvCombobox"
#pragma link "JvComponent"
#pragma link "JvDriveCtrls"
#pragma link "JvExControls"
#pragma link "JvExExtCtrls"
#pragma link "JvExMask"
#pragma link "JvExStdCtrls"
#pragma link "JvSpecialProgress"
#pragma link "JvToolEdit"
#pragma resource "*.dfm"
//---------------------------------------------------------------------------
TMainForm *MainForm = NULL;
extern TConfiguration *SelfImageConfig;
extern TProgramLog    *Log;
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Main form constructor
//
__fastcall TMainForm::TMainForm(TComponent* Owner) : TForm(Owner)
{
  DriveList = GetDriveList();
}
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// OnShow event handler - The form's objects are fully constructed but the
// form is not yet painted when this event occurs, so this is a great place
// to put code to initialize the initial state of the form's components.
//
void __fastcall TMainForm::FormShow(TObject *Sender)
{
  for (int n = 0; n < DriveList->Count; n++) {
    TDrive *Drive = (TDrive *)DriveList->Items[n];
    // cmbDriveList->ItemsEx->AddItem(Drive->DisplayName, 0, 0, 0, 0, Drive);
    cmbDriveList->AddItem(Drive->DisplayName, (TObject *)Drive);
  }  // for (int n = 0; n < DriveList->Count; n++)
  for (int n = 0; n < cmbDriveList->Items->Count; n++) {
    TDrive *Drive = (TDrive *)cmbDriveList->Items->Objects[n];
    if (Drive->MountPoint.Pos("C:"))
      cmbDriveList->ItemIndex = n;
  }  // for (int n = 0; n < cmbDriveList->Items->Count; n++)
  UpdateDriveInfo();
  edtTargetFile->InitialDir = SelfImageConfig->Values["Personal"];
}  // void __fastcall TMainForm::FormShow(TObject *Sender)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// cmbSourceDrive component OnChange event handler
//
void __fastcall TMainForm::cmbSourceDriveChange(TObject *Sender)
{
  UpdateDriveInfo();
}
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
void __fastcall TMainForm::edtTargetFileChange(TObject *Sender)
{
  btnOk->Enabled = !edtTargetFile->FileName.IsEmpty();
}
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Update the form's visual drive information components - called whenever
// the user changes source drives
//
void __fastcall TMainForm::UpdateDriveInfo(void)
{
  PARTITION_INFORMATION Partition;
  DISK_GEOMETRY Geometry;
  DWORD nCount;
  TDrive *Drive = (TDrive *)cmbDriveList->Items->Objects[cmbDriveList->ItemIndex];
  HANDLE hDrive;
  DWORD nSectorsPerCluster, nFreeClusters, nTotalClusters;

  try {
    if (!Drive->MountPoint.IsEmpty())
      hDrive = CreateFile(("\\\\.\\" + Drive->MountPoint.SubString(1,2)).c_str(), GENERIC_READ, FILE_SHARE_READ	| FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN | FILE_FLAG_BACKUP_SEMANTICS, NULL);
    else
      hDrive = XpCreateFile(Drive->DeviceName, GENERIC_READ, FILE_ATTRIBUTE_NORMAL, FILE_SHARE_READ | FILE_SHARE_WRITE, FILE_OPEN, FILE_SEQUENTIAL_ONLY);
    if (hDrive != INVALID_HANDLE_VALUE) {
      DeviceIoControl(hDrive, IOCTL_DISK_GET_PARTITION_INFO, NULL, 0, &Partition, sizeof(Partition), &nCount, NULL);
      DeviceIoControl(hDrive, IOCTL_DISK_GET_DRIVE_GEOMETRY	, NULL, 0, &Geometry, sizeof(Geometry), &nCount, NULL);
      CloseHandle(hDrive);
      if (Drive->DriveType == driveFloppy)
        nnTotalBytes = Geometry.Cylinders.QuadPart * Geometry.TracksPerCylinder * Geometry.SectorsPerTrack * Geometry.BytesPerSector;
      else
        nnTotalBytes = Partition.PartitionLength.QuadPart;
      nBytesPerSector = Geometry.BytesPerSector;
      nSectorsPerTrack = Geometry.SectorsPerTrack;
      lblSectorSizeValue->Caption  = nBytesPerSector;
      lblSectorCountValue->Caption = nnTotalBytes / nBytesPerSector;
      lblTotalSizeValue->Caption   = MakeByteLabel(nnTotalBytes);
    }  // if (hDrive != INVALID_HANDLE_VALUE)
  }  // try
  catch (...) {
    cmbDriveList->ItemIndex = -1;
  }  // catch
}  // void __fastcall TMainForm::UpdateDriveInfo(void)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
void __fastcall TMainForm::btnOkClick(TObject *Sender)
{
  HANDLE hInfile  = INVALID_HANDLE_VALUE;
  HANDLE hOutfile = INVALID_HANDLE_VALUE;
  OVERLAPPED Overlap;
  AnsiString sInFileDevice;
  void *buffer[2] = {NULL,NULL};
  unsigned nBufferSize;
  int nCurrent = 0;  // Current buffer
  int nPrevious = 1; // Previous buffer
  DWORD nBytesRead[2], nBytesWritten;
  DWORD nProcessMessagesFlag = 8;
  bool bWriteOk, bReadOk;
  TDrive *Drive;

  try {
    bCancelOperation = false;
    BytesWrittenTotal.QuadPart = 0;
    memset(&Overlap, 0, sizeof(Overlap));

    Drive = (TDrive *)cmbDriveList->Items->Objects[cmbDriveList->ItemIndex];
    if (Drive->DriveType == driveFixed)
      nBufferSize = nBytesPerSector * 8 * 64;
    else
      nBufferSize = nBytesPerSector * 36;
      
    if (Drive->MountPoint.IsEmpty())
      hInfile = XpCreateFile(Drive->DeviceName, GENERIC_READ | SYNCHRONIZE, FILE_ATTRIBUTE_NORMAL, FILE_SHARE_READ | FILE_SHARE_WRITE, FILE_OPEN, FILE_SYNCHRONOUS_IO_NONALERT | FILE_SEQUENTIAL_ONLY);
    else
      hInfile = CreateFile(("\\\\.\\" + Drive->MountPoint.SubString(1,2)).c_str(), GENERIC_READ, FILE_SHARE_READ	| FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN | FILE_FLAG_BACKUP_SEMANTICS, NULL);
    if (hInfile == INVALID_HANDLE_VALUE)
      throw ESelfImageFileError("SelfImage could not open drive " + Drive->DisplayName + ": '" + GetLastErrorMessage() +"'");

    buffer[0] = VirtualAlloc(NULL, nBufferSize, MEM_COMMIT, PAGE_READWRITE);
    if (!buffer[0])
      throw ESelfImageAllocError("SelfImage could not allocate memory for its internal disk read buffer: '" + GetLastErrorMessage() + "'");
    buffer[1] = VirtualAlloc(NULL, nBufferSize, MEM_COMMIT, PAGE_READWRITE);
    if (!buffer[1])
      throw ESelfImageAllocError("SelfImage could not allocate memory for its internal disk read buffer: '" + GetLastErrorMessage() + "'");

    hOutfile = CreateFile(edtTargetFile->FileName.c_str(), GENERIC_WRITE, 0, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED, NULL);
    if (hOutfile == INVALID_HANDLE_VALUE)
      throw ESelfImageFileError("SelfImage could not open the file " + edtTargetFile->FileName + " for output: '" + GetLastErrorMessage() + "'");

    SetFilePointer(hInfile, 0, NULL, FILE_BEGIN);
    SetFilePointer(hOutfile, 0, NULL, FILE_BEGIN);
    SetEndOfFile(hOutfile);

    VisualMode(modeOperating);
    UpdateStatus(true);

    if (!ReadFile(hInfile, buffer[nCurrent], nBufferSize, &nBytesRead[nCurrent], NULL))
      throw ESelfImageFileError("SelfImage encountered a problem reading from drive " + Drive->DisplayName + ": " + GetLastErrorMessage() );
    WriteFile(hOutfile, buffer[nCurrent], nBytesRead[nCurrent], NULL, &Overlap);
    BytesReadTotal.QuadPart = nBytesRead[nCurrent];
    nPrevious = nCurrent; nCurrent ^= 1;
    ReadLoop:
    while ( (bReadOk = ReadFile(hInfile, buffer[nCurrent], nBufferSize, &nBytesRead[nCurrent], NULL)) == true && nBytesRead[nCurrent] > 0 && !bCancelOperation) {
      BytesReadTotal.QuadPart += nBytesRead[nCurrent];
      bWriteOk = GetOverlappedResult(hOutfile, &Overlap, &nBytesWritten, TRUE);
      BytesWrittenTotal.QuadPart += nBytesWritten;
      Overlap.Offset = BytesWrittenTotal.LowPart;
      Overlap.OffsetHigh = BytesWrittenTotal.HighPart;
      if (!bWriteOk || nBytesWritten != nBytesRead[nPrevious])
        throw ESelfImageFileError("SelfImage encountered a problem writing to the file " + edtTargetFile->FileName + ".  The operation wrote " + nBytesWritten + "/" + nBytesRead[nPrevious] + "bytes (" + BytesWrittenTotal.QuadPart + "/" + nnTotalBytes + " total bytes written): '" + GetLastErrorMessage() + "'.");
      WriteFile(hOutfile, buffer[nCurrent], nBytesRead[nCurrent], NULL, &Overlap);
      nPrevious = nCurrent; nCurrent ^= 1;
      if (--nProcessMessagesFlag == 0) {
        nProcessMessagesFlag = 8;
        UpdateStatus();
        Application->ProcessMessages();
      }  // if (ProcessMessagesFlag == 0)
    }  // while (ReadFile(hInfile, buffer, nBufferSize, &nBytesRead, NULL) && nBytesRead > 0)
    // Devices opened with NtCreateFile() don't handle partial reads properly.  When a read fails, we have to step
    // down the read size to the sector size and try again.  This is a hack - there is probably a better way to do
    // this that involves understanding NtCreateFile() better.
    if (!bCancelOperation && nBufferSize > nBytesPerSector) {
      nBufferSize = nBytesPerSector;
      goto ReadLoop;
    }  // if (!bReadOk && nBufferSize > nBytesPerSector)
    if (!bCancelOperation) {
      bWriteOk = GetOverlappedResult(hOutfile, &Overlap, &nBytesWritten, TRUE);
      BytesWrittenTotal.QuadPart += nBytesWritten;
      if (!bWriteOk || nBytesWritten != nBytesRead[nPrevious])
        throw ESelfImageFileError("SelfImage encountered a problem writing to the file " + edtTargetFile->FileName + ".  The operation wrote " + nBytesWritten + "/" + nBytesRead[nPrevious] + "bytes (" + BytesWrittenTotal.QuadPart + "/" + nnTotalBytes + " total bytes written): '" + GetLastErrorMessage() + "'.");
    }  // if (!bCancelOperation)
/*  if (!bReadOk)  // This isn't working properly for devices opened with NtCreateFile()
      throw ESelfImageFileError("SelfImage encountered a problem reading from drive " + Drive->DisplayName + ": " + GetLastErrorMessage() ); */
    if (!bCancelOperation && BytesWrittenTotal.QuadPart != BytesReadTotal.QuadPart)
      throw ESelfImageFileError("SelfImage's internal counters of how much it read and how much it wrote don't match.");
    if (!bCancelOperation && BytesWrittenTotal.QuadPart < nnTotalBytes && !bReadOk)
      throw ESelfImageFileError("SelfImage encountered a problem reading from the device: " + GetLastErrorMessage());
    UpdateStatus(true);
  }  // try

  __finally {
    VisualMode(modeSetup);
    if (hOutfile != INVALID_HANDLE_VALUE)
      CloseHandle(hOutfile);
    if (buffer)
      VirtualFree(buffer, 0, MEM_RELEASE);
    if (hInfile != INVALID_HANDLE_VALUE)
      CloseHandle(hInfile);
  }  // __finally
}  // void __fastcall TMainForm::btnOkClick(TObject *Sender)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
void __fastcall TMainForm::btnCancelClick(TObject *Sender)
{
  bCancelOperation = true;
}
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
void __fastcall TMainForm::VisualMode(TMainFormVisualMode Mode)
{
  if (Mode == modeSetup) {
    btnOk->Enabled = !edtTargetFile->FileName.IsEmpty();
    btnCancel->Enabled = false;
    gbStatus->Visible = false;
    gbStatus->SendToBack();
  }  // if (Mode == modeSetup)

  if (Mode == modeOperating) {
    btnOk->Enabled = false;
    btnCancel->Enabled = true;
    gbStatus->Visible = true;
    gbStatus->BringToFront();
    ProgressBar->Maximum = ProgressBar->Width;
    ProgressBar->Position = 0;
    lblBytesTotalValue->Caption = lblTotalSizeValue->Caption;
    lblBytesWrittenValue->Caption = "0";
  }  // if (Mode == modeOperating)
  return;
}  // void __fastcall TMainForm::VisualMode(TMainFormVisualMode Mode)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
void __fastcall TMainForm::UpdateStatus(bool bInit)
{
  static LARGE_INTEGER StartTimer, LastTimer, PerformanceFrequency;
  static __int64 nnLastBytesWrittenTotal;
  LARGE_INTEGER Timer;
  long double ddTimeDifference, ddBytesDifference;

  if (bInit) {
    QueryPerformanceFrequency(&PerformanceFrequency);
    QueryPerformanceCounter(&StartTimer);
    LastTimer.QuadPart = StartTimer.QuadPart;
    nnLastBytesWrittenTotal = 0;
  } else {  // if (bInit)
    int nProgressTicks = BytesWrittenTotal.QuadPart * ProgressBar->Width / nnTotalBytes;
    ProgressBar->Position = nProgressTicks;
    lblBytesWrittenValue->Caption = MakeByteLabel(BytesWrittenTotal.QuadPart);
    QueryPerformanceCounter(&Timer);
    ddTimeDifference = Timer.QuadPart - LastTimer.QuadPart;
    ddTimeDifference /= PerformanceFrequency.QuadPart;
    if (ddTimeDifference > 1.0) {
      ddBytesDifference = BytesWrittenTotal.QuadPart - nnLastBytesWrittenTotal;
      lblWriteSpeedValue->Caption = MakeByteLabel(ddBytesDifference/ddTimeDifference) + "/s";
      ddBytesDifference = BytesWrittenTotal.QuadPart;
      ddTimeDifference = Timer.QuadPart - StartTimer.QuadPart;
      ddTimeDifference /= PerformanceFrequency.QuadPart;
      lblOverallSpeedValue->Caption = MakeByteLabel(ddBytesDifference/ddTimeDifference) + "/s";
      LastTimer.QuadPart = Timer.QuadPart;
      nnLastBytesWrittenTotal = BytesWrittenTotal.QuadPart;
    }  // if (dTimeDifference > 1.0)
  }  // if (bInit)
}  // void __fastcall StatusUpdateTimer(TObject *Sender)
//---------------------------------------------------------------------------


